"""Example datasets that are available for automatic downloading."""

AVAILABLE_CASES: dict[str, str] = {
    "example-input-volta": "10.4121/bbe10a2a-39dc-4098-a69f-0f677d06ecdd.v2",
    "example-input-eiffel": "10.4121/f9572240-f179-4338-9e1b-82c5598529e2.v1",
}



"""Routines specific for preprocessing pressure level data."""
import numpy as np
import xarray as xr

from wam2layers.config import Config
from wam2layers.preprocessing.utils import interpolate, sortby_ndarray


def extend_pressurelevels(
    level_data: xr.Dataset,
    surface_data: xr.Dataset,
    config: Config,
) -> tuple[xr.Dataset, xr.DataArray]:
    """Extend the levels to top of atmosphere, surface and 2-layer boundary level.

    The input data does not have data at the surface and
    and the top of atmosphere. The input data is extended to
    these edges.

    For proper integration, extra layers have to be inserted
    near the 2-layer boundary. This boundary level is hard
    coded as a function of surface pressure and not customizable.

    Args:
        level_data: Dataset with the following variables:
            q: Specific humidity at pressure levels
            u: Eastward horizontal wind speed at pressure levels
            v: Northward horizontal wind speed at pressure levels
        surface_data: Dataset with the following variables:
            ps: Air pressure at the surface
            qs: Specific humidity at the surface
            us: Eastward horizontal wind speed at the surface
            vs: Northward horizontal wind speed at the surface

    Returns:
        Dataset with the following variables (interpolated to new pressure levels):
            p, q, u, v
        The pressure at the division between the two layers
    """
    level_ax: int = level_data["u"].get_axis_num("level")  # type: ignore

    da_p = level_data["level"].broadcast_like(level_data["u"])

    p = da_p.to_numpy()
    u = level_data["u"].to_numpy()
    v = level_data["v"].to_numpy()
    q = level_data["q"].to_numpy()

    # Insert top of atmosphere values
    # Assume wind at top same as values at lowest pressure, humidity at top = 0
    p0 = p.argmin(axis=level_ax).flatten()[0]
    p = np.insert(p, 0, 0.0, axis=level_ax)
    q = np.insert(q, 0, 0.0, axis=level_ax)
    v = np.insert(v, 0, np.take(v, p0, axis=level_ax), axis=level_ax)
    u = np.insert(u, 0, np.take(u, p0, axis=level_ax), axis=level_ax)

    # Insert surface level values
    p = np.insert(p, 0, surface_data["ps"], axis=level_ax)
    q = np.insert(q, 0, surface_data["qs"], axis=level_ax)
    v = np.insert(v, 0, surface_data["vs"], axis=level_ax)
    u = np.insert(u, 0, surface_data["us"], axis=level_ax)

    # Sort arrays by pressure (ascending)
    u = sortby_ndarray(u, p, axis=level_ax)
    v = sortby_ndarray(v, p, axis=level_ax)
    q = sortby_ndarray(q, p, axis=level_ax)
    p = sortby_ndarray(p, p, axis=level_ax)

    # Insert boundary level values (at a dummy pressure value)
    p_boundary = (
        config.pressure_boundary_factor * np.array(surface_data["ps"])
        + config.pressure_boundary_offset
    )
    u = np.insert(u, 0, interpolate(p_boundary, p, u, axis=level_ax), axis=level_ax)
    v = np.insert(v, 0, interpolate(p_boundary, p, v, axis=level_ax), axis=level_ax)
    q = np.insert(q, 0, interpolate(p_boundary, p, q, axis=level_ax), axis=level_ax)
    p = np.insert(p, 0, p_boundary, axis=level_ax)

    # Sort arrays by pressure once more (ascending)
    u = sortby_ndarray(u, p, axis=level_ax)
    v = sortby_ndarray(v, p, axis=level_ax)
    q = sortby_ndarray(q, p, axis=level_ax)
    p = sortby_ndarray(p, p, axis=level_ax)

    # Reset level coordinate as its values have become meaningless
    nlev = np.size(u, axis=level_ax)
    levs = np.arange(nlev)

    # reconstruct dataarrays
    coords = dict(level_data["u"].coords)
    coords["level"] = levs

    output = xr.Dataset()  # New dataset as the coordinate size has changed
    output["u"] = xr.DataArray(u, coords, level_data["u"].dims)
    output["v"] = xr.DataArray(v, coords, level_data["v"].dims)
    output["q"] = xr.DataArray(q, coords, level_data["q"].dims)
    output["p"] = xr.DataArray(p, coords, level_data["u"].dims)
    pb = xr.DataArray(p_boundary, surface_data["ps"].coords, surface_data["ps"].dims)

    return output, pb


def interp_dp_midpoints(
    level_data: xr.Dataset,
) -> tuple[xr.Dataset, xr.DataArray]:
    """Interpolate the data to midpoints to allow for integration to two layers.

    Args:
        level_data: Dataset with the following variables:
            q: Specific humidity at levels
            u: Eastward horizontal wind speed at levels
            v: Northward horizontal wind speed at levels
            p: Air pressure at levels

    Returns:
        Dataset with the following variables:
            p interpolated to midpoints of pressure levels,
            q interpolated to midpoints of pressure levels,
            u interpolated to midpoints of pressure levels,
            v interpolated to midpoints of pressure levels,
        Pressure difference between the pressure level bounds,
    """

    # Calculate pressure jump
    dp = level_data["p"].diff("level")
    assert np.all(dp >= 0), "Pressure levels should increase monotonically"

    # Interpolate to midpoints
    midpoints = 0.5 * (
        level_data["level"].to_numpy()[1:] + level_data["level"].to_numpy()[:-1]
    )
    dp = dp.assign_coords(level=midpoints)
    interped_data = level_data.interp(level=midpoints)
    return interped_data, dp


def mask_below_surface_data(
    level_data: xr.Dataset,
    ps: xr.DataArray,
) -> tuple[xr.Dataset, xr.DataArray]:
    """Mask any points that fall below the surface.

    Args:
        level_data: Dataset with the following variables:
            q: Specific humidity at levels
            u: Eastward horizontal wind speed at levels
            v: Northward horizontal wind speed at levels
            p: Air pressure at levels
        ps: Air pressure at the surface

    Returns:
        The input data, with the values below the surface (based on surface
            air pressure) masked out.
    """

    # mask values below surface
    above_surface = level_data["p"] < ps.broadcast_like(level_data["p"])
    return level_data.where(above_surface)



"""Functions for downloading example cases."""

import logging
import re
import shutil
from pathlib import Path

import requests
import threddsclient

logger = logging.getLogger(__name__)


def parse_doi(doi_string):
    """Extract prefix and suffix from DOI.

    Structure of DOIs:

        <resolver>/<prefix>/<suffix>

    For example:

        https://doi.org/10.4121/f9572240-f179-4338-9e1b-82c5598529e2.v1

    Here the prefix is `10.4121` is specific for 4TU. The suffix (generated by
    4TU) is `f9572240-f179-4338-9e1b-82c5598529e2.v1`

    Note that the suffix may contain slashes, and 4TU inserts a version in it,
    but this is not part of the official doi spec.

    For more info, see
    https://www.crossref.org/documentation/member-setup/constructing-your-dois/
    """
    regex = r"(?P<prefix>10.\d{4,9})\/(?P<suffix>[-._;()\/:A-Z0-9]+$)"
    matches = re.search(regex, doi_string, re.IGNORECASE)

    if matches:
        # Extract the capturing groups
        prefix = matches.group("prefix")
        suffix = matches.group("suffix")

        return prefix, suffix
    else:
        raise ValueError(f"Unable to parse DOI {doi_string}.")


def create_target_directory(name):
    """Create the target directory but prevent overwrites."""
    target_directory = Path.cwd() / name
    logger.info(f"Creating directory {target_directory}.")

    try:
        target_directory.mkdir(exist_ok=False)
    except FileExistsError as exc:
        raise FileExistsError(
            f"Target directory {name} exists. Stopping download to prevent overwriting your files. "
            "If you want to replace your files, first (re)move the directory manually."
        ) from exc

    return target_directory


def download(url, destination, local_filename=None):
    """Download file from url to destination."""
    if local_filename is None:
        local_filename = url.split("/")[-1]

    local_path = destination / local_filename

    logging.info(f"Downloading {local_filename}")
    with requests.get(url, stream=True) as r:
        with open(local_path, "wb") as f:
            shutil.copyfileobj(r.raw, f)


def get_4tu_id(doi):
    """Extract 4TU ID from DOI.

    As far as I can tell, DOI is constructed like so:
    10.4121/<4TU-ID>.v<version>

    Where 10.4121 is the prefix for 4TU.
    """
    _, suffix = parse_doi(doi)

    # Get UUID from DOI
    id_4tu = suffix.split(".")[0]

    return id_4tu


def get_4tu_catalog(uuid):
    """Retrieve the catalogue from 4TU API.

    API is undocumented, but similar to figshare.
    Catalog is available via articles endpoint, under
    custom_fields -> Data Link
    """
    response = requests.get(f"https://data.4tu.nl/v2/articles/{uuid}")
    response.raise_for_status()
    dataset_info = response.json()

    for field in dataset_info["custom_fields"]:
        if field["name"] == "Data Link":
            catalog_url_html = field["value"][0]

    catalog_url_xml = catalog_url_html.replace(".html", ".xml")
    return catalog_url_xml


def get_4tu_files(uuid):
    """Retrieve file names and download links for files.

    This concerns any (additional) files that are not available via OpenDAP
    (e.g. config.yaml).
    """
    response = requests.get(f"https://data.4tu.nl/v2/articles/{uuid}")
    response.raise_for_status()
    dataset_info = response.json()

    files = {}
    for file_info in dataset_info["files"]:
        name = file_info["name"]
        url = file_info["download_url"]
        files[name] = url

    return files


def download_from_4tu(doi, name):
    """Download each file using threddsclient as crawler.

    Args:
        doi: doi of the dataset (with or without https://doi.org/)
        name: human-readable name for the dataset, used to create target directory
    """
    logger.info(f"Downloading 4TU dataset to ./{name}")

    target_dir = create_target_directory(name)
    id_4tu = get_4tu_id(doi)
    catalog_url = get_4tu_catalog(id_4tu)

    # Extract individual file urls from catalogue and download them separately
    file_urls = threddsclient.download_urls(catalog_url)
    for file_url in file_urls:
        download(file_url, target_dir)

    # Extract additional files from file listing
    additional_files = get_4tu_files(id_4tu)
    for filename, url in additional_files.items():
        download(url, target_dir, filename)


def download_from_doi(doi, name):
    """Parse doi and redirect to provider-specific download function.

    Args:
        doi: doi of the dataset (with or without https://doi.org/)
        name: human-readable name for the dataset, used to create target directory
    """
    prefix, _ = parse_doi(doi)

    if prefix == "10.4121":  # 4TU
        download_from_4tu(doi, name)
    elif prefix == "10.5281":  # Zenodo
        raise NotImplementedError(
            "Download utility for Zenodo has not been implemented."
        )
    else:
        raise ValueError(f"Unknown provider with doi prefix {prefix}, cannot download.")



"""Loading ERA5 data for preprocessing."""
from datetime import datetime as pydt
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from wam2layers import __version__
from wam2layers.config import Config
from wam2layers.preprocessing import logger
from wam2layers.preprocessing.utils import (
    accumulation_to_flux,
    calculate_humidity,
    log_once,
    midpoints,
)

PREPROCESS_ATTRS = {
    "title": "ERA5 data preprocessed for use in WAM2layers",
    "history": (
        f"created on {pydt.utcnow():%Y-%m-%dT%H:%M:%SZ} "
        f"using wam2layers version {__version__}."
    ),
    "source": (
        "ECMWF Reanalysis v5 (ERA5), "
        "www.ecmwf.int/en/forecasts/dataset/ecmwf-reanalysis-v5"
    ),
    "references": "doi.org/10.5281/zenodo.7010594, doi.org/10.5194/esd-5-471-2014",
    "Conventions": "CF-1.6",
}


def get_input_data(datetime: pd.Timestamp, config: Config):
    """Get the preprocessing input data for ERA5."""
    data = xr.Dataset()
    data["q"] = load_data("q", datetime, config)  # in kg kg-1
    data["u"] = load_data("u", datetime, config)  # in m/s
    data["v"] = load_data("v", datetime, config)  # in m/s
    data["ps"] = load_data("sp", datetime, config)  # in Pa
    if config.level_type == "pressure_levels":
        d2m = load_data("d2m", datetime, config)  # Dew point in K
        data["qs"] = calculate_humidity(d2m, data["ps"])  # kg kg-1
        data["us"] = load_data("u10", datetime, config)  # in m/s
        data["vs"] = load_data("v10", datetime, config)  # in m/s
    if config.level_type == "model_levels":
        data["dp"] = get_dp_modellevels(data["ps"], config.levels)

    data["precip"], data["evap"] = preprocess_precip_and_evap(datetime, config)
    try:
        data["tcw"] = load_data("tcw", datetime, config)
    except FileNotFoundError:
        log_once(
            logger,
            "Total column water input data not available; using water vapour only",
        )

    if config.level_type == "pressure_levels":
        data["level"] = data["level"] * 100  # hPa --> Pa
        data["level"].attrs.update(units="Pa")

    return data, PREPROCESS_ATTRS


def load_data(variable: str, datetime: pd.Timestamp, config: Config) -> xr.DataArray:
    """Load data for given variable and date."""
    template = config.filename_template

    # If it's a 4d variable, we need to set the level type
    if variable in ["u", "v", "q"]:
        if config.level_type == "model_levels":
            prefix = "_ml"
        elif config.level_type == "pressure_levels":
            prefix = "_pl"
    # If it's a 3d variable we do not need to set the level type
    else:
        prefix = ""

    # Load data
    filepath = template.format(
        year=datetime.year,
        month=datetime.month,
        day=datetime.day,
        hour=datetime.hour,
        minute=datetime.minute,
        levtype=prefix,
        variable=variable,
    )
    da = xr.open_dataarray(filepath).sel(time=datetime.strftime("%Y%m%d%H%M"))

    if "lev" in da.coords:
        da = da.rename(lev="level")

    # If it's 4d data we want to select a subset of the levels
    if variable in ["u", "v", "q"] and isinstance(config.levels, list):
        return da.sel(level=config.levels)

    return da


def preprocess_precip_and_evap(
    datetime: pd.Timestamp, config: Config
) -> tuple[xr.DataArray, xr.DataArray]:
    """Load and pre-process precipitation and evaporation."""
    # All incoming units are accumulations (in m) since previous time step
    evap = load_data("e", datetime, config)
    precip = load_data("tp", datetime, config)  # total precipitation

    # Transfer negative (originally positive) values of evap to precip
    precip = np.maximum(precip, 0) + np.maximum(evap, 0)
    precip = precip

    # Change sign convention to all positive,
    evap = np.abs(np.minimum(evap, 0))

    precip = accumulation_to_flux(precip, input_frequency=config.input_frequency)
    evap = accumulation_to_flux(evap, input_frequency=config.input_frequency)
    return precip, evap


def load_era5_ab():
    """Load a and b coefficients."""
    table = Path(__file__).parent / "tableERA5model_to_pressure.csv"
    df = pd.read_csv(table)
    a = df["a [Pa]"].values
    b = df.b.values
    return a, b


def get_edges(era5_modellevels):
    """Get the values of a and b at the edges of a subset of ERA5 modellevels."""
    a, b = load_era5_ab()

    if era5_modellevels == "all":
        return a, b

    # With python indexing starting at 0 and 137 full levels,
    # count from 0-136 instead of 1-137.
    era5_modellevels = [i - 1 for i in era5_modellevels]

    # Calculate the values of a and be at midpoints and extract levels
    a_full = midpoints(a)[era5_modellevels]
    b_full = midpoints(b)[era5_modellevels]

    # Calculate the values of a and b at the edges *between* these levels
    a_edge = midpoints(a_full)
    b_edge = midpoints(b_full)

    # Reinsert the original outer edges
    a_edge = np.block([a[0], a_edge, a[-1]])
    b_edge = np.block([b[0], b_edge, b[-1]])

    return a_edge, b_edge


def get_dp_modellevels(sp, levels):
    """Calculate pressure jump over subset of era5 modellevels."""
    a, b = get_edges(levels)

    # Use dummy level -1 to avoid collision with existing levels
    # first coord will be discarded anyway on taking the diff below.
    a = xr.DataArray(a, coords={"level": np.insert(levels, 0, -1)})
    b = xr.DataArray(b, coords={"level": np.insert(levels, 0, -1)})

    p_edge = a + b * sp

    # Calculate the difference between the pressure levels
    dp = p_edge.diff("level")  # in Pa

    return dp



import numpy as np
import xarray as xr

from wam2layers.utils.grid import get_boundary


def test_get_boundary():
    a = np.arange(12).reshape(3, 4)
    a_xr = xr.DataArray(a)
    result = get_boundary(a_xr)
    expected = np.array(
        [
            [True, True, True, True],
            [True, False, False, True],
            [True, True, True, True],
        ]
    )
    np.testing.assert_array_equal(result, expected)


def test_get_boundary_periodic():
    a = np.arange(12).reshape(3, 4)
    a_xr = xr.DataArray(a)
    result = get_boundary(a_xr, periodic=True)
    expected = np.array(
        [
            [True, True, True, True],
            [False, False, False, False],
            [True, True, True, True],
        ]
    )
    np.testing.assert_array_equal(result, expected)



"""Preprocessing input data validation."""
import numpy as np
import xarray as xr

from wam2layers.config import Config


class ValidationError(Exception):
    ...


def validate_input(data: xr.Dataset, config: Config) -> None:
    validate_dims(data)
    validate_vars(data, config)
    check_nans(
        data
    )  # note: needs to load all data in memory... move to pre-output writing


def validate_dims(data: xr.Dataset) -> None:
    expected_dims = ["latitude", "longitude", "level"]
    for dim in expected_dims:
        if dim not in data.dims:
            msg = (
                f"Dimension '{dim}' is missing from the input data.\n"
                f" Dims present in data: {data.dims}."
            )
            raise ValidationError(msg)


def validate_vars(data: xr.Dataset, config: Config) -> None:
    expected_level_vars = ["q", "u", "v"]
    expected_surface_vars = ["ps", "precip", "evap"]
    if config.level_type == "pressure_levels":
        expected_surface_vars.extend(["qs", "us", "vs"])
    if config.level_type == "model_levels":
        expected_level_vars.extend(["dp"])

    # Check if the variables are present
    for var in expected_level_vars + expected_surface_vars:
        if var not in data.data_vars:
            msg = (
                f"Variable '{var}' is missing from the input data.\n"
                f" Variables present in data: {data.data_vars}."
            )
            raise ValidationError(msg)

    # Make sure surface variables have correct dims
    for var in expected_surface_vars:
        if set(data[var].dims) != {"latitude", "longitude"}:
            msg = (
                "Surface variables should only have the dimensions latitude, "
                "longitude.\n"
                f" Var {var} has dimensions: {data[var].dims}."
            )
            raise ValidationError(msg)

    for var in expected_level_vars:
        if set(data[var].dims) != {"level", "latitude", "longitude"}:
            msg = (
                "Pressure/model Level variables should have the dimensions level, "
                "latitude, longitude.\n"
                f" Var {var} has dimensions: {data[var].dims}."
            )
            raise ValidationError(msg)


def check_nans(data: xr.Dataset):
    for var in data.data_vars:
        nans = data[var].isnull()
        if nans.any():
            msg = f"{np.sum(nans)} NaN values found in variable '{var}'"
            raise ValidationError(msg)



"""General preprocessing routine."""
import pandas as pd
import xarray as xr

from wam2layers.config import Config
from wam2layers.preprocessing import era5, logger
from wam2layers.preprocessing.input_validation import validate_input
from wam2layers.preprocessing.pressure_levels import (
    extend_pressurelevels,
    interp_dp_midpoints,
    mask_below_surface_data,
)
from wam2layers.preprocessing.utils import add_bounds
from wam2layers.preprocessing.xarray_append import append_to_netcdf
from wam2layers.reference.variables import PREPROCESSED_DATA_ATTRIBUTES


def get_input_dates(config: Config) -> pd.DatetimeIndex:
    """Dates for pre-processing."""
    return pd.date_range(
        start=config.preprocess_start_date,
        end=config.preprocess_end_date,
        freq=config.input_frequency,
    )


def get_input_data(
    datetime: pd.Timestamp, config: Config, source: str
) -> tuple[xr.Dataset, dict[str, str]]:
    """Retrieve input data from a specified source (e.g. ERA5, CMIP6).

    This will load the xr.Dataset required for the preprocessing routines.
        This dataset has the following coordinates: latitude, longitude, level.
            Note that for pressure level data, the level coordinates should be the air
            pressure (Pa). For model level data these are integers, counting down
            towards the surface.
        It has the following variables:
            q: Specific humidity at pressure levels (kg/kg)
            u: Eastward horizontal wind speed at pressure levels (m/s)
            v: Northward horizontal wind speed at pressure levels (m/s)
            ps: Air pressure at the surface (Pa)
            twc (optional): Total column water content (kg/m2)
        If the data uses model layers, this dataset has the variable:
            dp: Pressure difference over each model layer (Pa).
        If the data uses pressure layers, the dataset will have the following additional
        variables:
            qs: Specific humidity at the surface (kg/kg)
            us: Eastward horizontal wind speed at the surface (m/s)
            vs: Northward horizontal wind speed at the surface (m/s)

    Args:
        datetime: Which day of data should be loaded
        config: WAM2layers configuration
        source (optional): Which source to use. Defaults to "ERA5".

    Returns:
        Prepared input data
        Attributes corresponding to the input data source
    """
    if source == "ERA5":
        data, input_data_attrs = era5.get_input_data(datetime, config)
    else:
        msg = "Only the ERA5 input data loader has been implemented."
        raise NotImplementedError(msg)

    validate_input(data, config)
    return data, input_data_attrs


def prep_experiment(config_file: Config, data_source: str):
    """Pre-process all data for a given config file.

    This function expects the following configuration settings:

    - preprocess_start_date: formatted as YYYYMMDD, e.g. '20210701'
    - preprocess_end_date: formatted as YYYYMMDD, e.g. '20210716'
    - level_type: either "pressure_levels" or "model_levels"
    - levels: "all" or a list of integers with the desired (model or pressure)
      levels.
    - input_folder: path where raw era5 input data can be found, e.g.
      /home/peter/WAM2layers/era5_2021
    - preprocessed_data_folder: path where preprocessed data should be stored.
      This directory will be created if it does not exist. E.g.
      /home/peter/WAM2layers/preprocessed_data_2021
    - filename_prefix: Fixed part of filename. This function will infer the
      variable name and add _ml for model level data. E.g. with prefix =
      "FloodCase_202107" this function will be able to find
      FloodCase_202107_ml_u.nc or FloodCase_202107_u.nc and
      FloodCase_202107_sp.nc
    """
    config = Config.from_yaml(config_file)

    for datetime in get_input_dates(config):
        is_new_day = datetime == datetime.floor("1d")
        logger.info(datetime)

        input_data, input_data_attrs = get_input_data(datetime, config, data_source)
        surface_data = input_data.drop_dims("level")
        level_data = input_data.drop_vars(surface_data.data_vars)

        if config.level_type == "pressure_levels":
            level_data, pb = extend_pressurelevels(level_data, surface_data, config)
            level_data, dp = interp_dp_midpoints(level_data)
            level_data = mask_below_surface_data(level_data, surface_data["ps"])
        else:
            dp = level_data["dp"]

        # Calculate column water vapour
        g = 9.80665  # gravitational acceleration [m/s2]
        cwv = level_data["q"] * dp / g  # (kg/m2)

        if "tcw" in surface_data:
            # Calculate column water instead of column water vapour
            correction = surface_data["tcw"] / cwv.sum(dim="level")
            cw = correction * cwv  # column water (kg/m2)
            if is_new_day:
                logger.info(
                    "Total column water correction:\n"
                    "    ratio total column water / computed integrated water vapour\n"
                    f"    mean over grid for this timestep {correction.mean().item():.4f}"
                )
        else:  # Fluxes will be calculated based on the column water vapour
            cw = cwv

        # Integrate fluxes and states to upper and lower layer
        if config.level_type == "model_levels":
            lower_layer = dp["level"] > config.level_layer_boundary
            upper_layer = ~lower_layer

        if config.level_type == "pressure_levels":
            upper_layer = level_data["p"] < pb.broadcast_like(level_data["p"])
            lower_layer = ~upper_layer

        # Vertically integrate state over two layers
        s_lower = cw.where(lower_layer).sum(dim="level")
        s_upper = cw.where(upper_layer).sum(dim="level")

        # Determine the fluxes
        fx = level_data["u"] * cw  # eastward atmospheric moisture flux (kg m-1 s-1)
        fy = level_data["v"] * cw  # northward atmospheric moisture flux (kg m-1 s-1)

        # Vertically integrate fluxes over two layers
        fx_lower = fx.where(lower_layer).sum(dim="level")  # kg m-1 s-1
        fy_lower = fy.where(lower_layer).sum(dim="level")  # kg m-1 s-1
        fx_upper = fx.where(upper_layer).sum(dim="level")  # kg m-1 s-1
        fy_upper = fy.where(upper_layer).sum(dim="level")  # kg m-1 s-1

        # Combine everything into one dataset
        ds = (
            xr.Dataset(
                {
                    "fx_upper": fx_upper.assign_attrs(units="kg m-1 s-1"),
                    "fy_upper": fy_upper.assign_attrs(units="kg m-1 s-1"),
                    "fx_lower": fx_lower.assign_attrs(units="kg m-1 s-1"),
                    "fy_lower": fy_lower.assign_attrs(units="kg m-1 s-1"),
                    "s_upper": s_upper.assign_attrs(units="kg m-2"),
                    "s_lower": s_lower.assign_attrs(units="kg m-2"),
                    "evap": surface_data["evap"],
                    "precip": surface_data["precip"],
                }
            )
            .expand_dims("time")
            .astype("float32")
        )
        add_bounds(ds)

        # Add attributes
        for var in PREPROCESSED_DATA_ATTRIBUTES:
            ds[var].attrs.update(PREPROCESSED_DATA_ATTRIBUTES[var])
        ds.attrs.update(input_data_attrs)

        # Save preprocessed data
        filename = f"{datetime.strftime('%Y-%m-%d')}_fluxes_storages.nc"
        output_path = config.preprocessed_data_folder / filename

        if is_new_day:
            comp = dict(zlib=True, complevel=8)
            encoding = {var: comp for var in ds.data_vars}
            time_encoding = {"units": "seconds since 1900-01-01"}
            encoding["time"] = time_encoding
            ds.to_netcdf(
                output_path, unlimited_dims=["time"], mode="w", encoding=encoding
            )

        else:
            append_to_netcdf(output_path, ds, expanding_dim="time")
